<?xml version="1.0"?>

<st-source>
<!-- Changes in change set FRMI-503285 -->
<time-stamp>From VisualWorks®, 7.4.1 of May 30, 2006 on June 17, 2008 at 4:09:14 pm</time-stamp>


<methods>
<class-id>PROVISO.FormulaInstallerApplication</class-id> <category>hooks</category>

<body package="PV-CME-Core" selector="hookStart">hookStart

	| out result |
	out := OS.Stdout.
	super hookStart.
	self dbSession connect.
	self dbSession isConnected ifFalse: 
		[out nextPutAll: 'Unable to establish DB connection, exiting.'; cr.
		^self stopImmediate].
	self dbSession begin.
	result := [
		self formulaDocuments
			do: [ :doc |
				OS.Stdout 
					nextPutAll: 'Processing ';
					nextPutAll: (doc asFilename asString);
					cr.
				FormulaSet 
					installFromFile: doc asFilename 
					infoOn: out]
			separatedBy: [OS.Stdout cr].
		true]
	on: Error do: [:signal |
		self logError: signal description.
		OS.Stdout 
			nextPutAll: 'Error: ';
			nextPutAll: signal description;
			cr.
		signal return: false].

	result
		ifTrue: [self dbSession commit]
		ifFalse: [self dbSession rollback].
	self stopImmediate.</body>
</methods>


<methods>
<class-id>PROVISO.FormulaSet class</class-id> <category>instance creation</category>

<body package="PV-ES-Formulas" selector="installFromFile:infoOn:">installFromFile: aFilename infoOn: aStrm 

	| fileSet dbSet deltas model argMapInsert highWatermark |
	highWatermark := self timeFacade nowAsSeconds.
	(self 
		validateFromFile: aFilename
		infoOn: aStrm) 
			ifFalse: [^nil].
	PROVISO.UBA.Environment new do: 
		[model := 
			PvModel new
				initializeForCme;
				initIdManager;
				yourself.
		"Resync up to 1 day to avoid any issues with local timezone settings"
		model 
			resyncDomainManagers: #(formulaManager calendarManager)
			collectors: nil
			between: 0
			and: self timeFacade now asUTCSeconds + 8640.
		fileSet := self loadFromFile: aFilename.
		fileSet do: [:ea | ea resyncHistoricalDate: Infinity positive].
		dbSet := 
			self 
				formulasForDoc: fileSet documentName
				inModel: model.
		deltas := fileSet asMutationsOf: dbSet.
		"TODO: only store the document if there has been a change in the text of the document...also, need to be able to handle document include statements."
		model formulaManager 
			storeDocument: aFilename
			named: fileSet documentName.
		"The following code needs to be retained and will be utilized in the future, for now, the plumbing is not in place to support it -spair/apikovsky"
		"argMapInsert := (String new: 1024) streamContentsDo: [:stream |
					stream nextPutAll: 'insert into FRML_ARG (idx_ind, str_arg_path, dte_date) values ('; cr.
					(model formulaManager argumentMap associations 
						select: [:eachAssoc | eachAssoc value key &gt;= highWatermark])
							do: [:eachAssoc |
								stream
									print: eachAssoc value value; nextPut: $,; space;
									nextPutAll: eachAssoc key; nextPut: $,; space;
									print: eachAssoc value key]
							separatedBy: [stream  nextPut: $,; cr].
					stream cr; nextPutAll: ')'.
				].
				self dbSession executeNonSelect: argMapInsert."
		aStrm nextPutAll: 'Deleted Formulas:'.
		(deltas at: #deletes) isEmpty 
			ifTrue: [aStrm nextPutAll: ' (none)'].
		aStrm cr.
		(deltas at: #deletes) do: 
			[:formulaName | 
			aStrm
				nextPutAll: '   ';
				nextPutAll: formulaName;
				cr.
			(dbSet formulaNamed: formulaName) dbDelete].
		aStrm nextPutAll: 'Inserted Formulas:'.
		(deltas at: #inserts) isEmpty 
			ifTrue: [aStrm nextPutAll: ' (none)'].
		aStrm cr.
		(deltas at: #inserts) do: 
			[:formulaName | 
			| formula |
			aStrm
				nextPutAll: '   ';
				nextPutAll: formulaName;
				cr.
			formula := fileSet formulaNamed: formulaName.
			model formulaManager 
				dbAddFormula: formula
				named: formulaName
				withSeparator: $~.
			model formulaManager 
				associateFormula: formula
				withDocument: fileSet documentName].
		aStrm nextPutAll: 'Updated Formulas:'.
		(deltas at: #updates) addAll: (deltas at: #unchanged).
		"update all formulas, even the unchanged ones"
		(deltas at: #updates) isEmpty 
			ifTrue: [aStrm nextPutAll: ' (none)'].
		aStrm cr.
		(deltas at: #updates) do: 
			[:formulaName | 
			| formula |
			aStrm
				nextPutAll: '   ';
				nextPutAll: formulaName;
				cr.
			formula := fileSet formulaNamed: formulaName.
			formula id: (dbSet formulaNamed: formulaName) id.
			(dbSet formulaNamed: formulaName) group ifNotNil: [:grp | formula group: grp].
			formula dbUpdate]].

	self installDummyFormulasForCalendarsIn: model infoOn: aStrm</body>
</methods>



</st-source>
