<?xml version="1.0"?>

<st-source>
<!-- Changes in change set PV -->
<time-stamp>From VisualWorks®, 7.4.1 of May 30, 2006 on July 1, 2008 at 1:01:38 pm</time-stamp>


<methods>
<class-id>PROVISO.PvModelResyncAgent</class-id> <category>resync actions</category>

<body package="PV-DM-Managers" selector="updatedProperty:">updatedProperty: property

	updatedProperties ifNotNil: [updatedProperties add: property]</body>
</methods>

<methods>
<class-id>PROVISO.PvModelResyncAgent</class-id> <category>initialize - release</category>

<body package="PV-DM-Managers" selector="initialize">initialize

	super initialize.
	self removedObjects: (PvDictionary newSizedFor: 1000).
	bufferActions := true.
	lastStartSeconds := lastEndSeconds := 0.
	self resyncActions: OrderedCollection new; testResyncActions: OrderedCollection new.
	self updatedProperties: OrderedCollection new.</body>
</methods>


<methods>
<class-id>PROVISO.RecomputationEngine</class-id> <category>As yet unclassified</category>

<body package="PV-DC-Recomputations" selector="computeMetadataChangeForMetricId:resourceId:between:and:">computeMetadataChangeForMetricId: metricId resourceId: resourceId between: startTime and: endTime 
	"This method needs to figure out what metadata has changed for the given mid / rid, in the given scope.
	Non-input metadata changes can include formula definition changes, and subelement changes."
	
	|  formula changes |
	changes := OrderedCollection new.
	formula := pvModel formulaManager itemWithId: metricId.
	self dbSession 
		readObjectsOfClass: PvFormula
		with: 
			(PvFormula selectStatementWithParameters: 
				#startSeconds -&gt; startTime , (#endSeconds -&gt; endTime) , 
					(#id -&gt; metricId))
		into: 
			[:each | 
			changes add: 
				(RecomputationDbChangeWrapper 
					on: 
						(self 
							buildUpdateBlockForFormula: formula
							from: each)
					from: each timestamp
					to: endTime)].
	self dbSession 
		readObjectsOfClass: PvCmeSubelement
		with: 
			(PvCmeSubelement selectStatementWithParameters: 
				#startSeconds -&gt; startTime , (#endSeconds -&gt; endTime) , 
					(#id -&gt; resourceId))
		into: 
			[:each | 
			changes add: 
				(RecomputationDbChangeWrapper 
					on: (self buildUpdateBlockForSubelement: each)
					from: each timestamp
					to: endTime)].
	^changes</body>

<body package="PV-DC-Recomputations" selector="recompile:withFormula:">recompile: anOperator withFormula: aFormula 
	" Assume aTimeSeries is a Smee formula-based time series.
	Replace the prototype with the new formula and reshape the inputs "
	
	| oldInputs newInputs obsoleteInputs |
	oldInputs := anOperator inputs.
	newInputs := Array new: aFormula arguments size.
	" Get all the inputs that are preserved between in the new version of the formula and put them in proper place at the new inputs array"
	aFormula arguments keysAndValuesDo: 
		[:ix :eachArgument | 
		| oldInput |
		oldInput := 
			oldInputs 
				detect: [:eachInput | eachInput notNil and: [eachInput sequenceNumber = eachArgument id]]
				ifNone: [nil].
		oldInput ifNotNil: [newInputs at: ix put: oldInput]].

	" These inputs should have been deleted in an orderly fashion, so their presence may be an indication of some problem in the logic "
	obsoleteInputs := 
		oldInputs reject: 
			[:eachInput | 
			eachInput isNil 
				or: [aFormula arguments anySatisfy: [:eachArgument | eachInput sequenceNumber = eachArgument id]]].
	obsoleteInputs do: 
		[:eachInput | 
		self system 
			logWarning: #DISCONNECT_OBSOLETE_INPUT
			messageId: nil
			text: 
				'Disconnecting obsolete input on formula reshape. Time series ' , 
					anOperator printString , ', input: ' , eachInput printString.
		eachInput disconnect].
	anOperator smeePrototype: aFormula.
	anOperator inputs: newInputs</body>
</methods>


<methods>
<class-id>PROVISO.RecomputationEngine</class-id> <category>private</category>

<body package="PV-DC-Recomputations" selector="doUpdateForComplexFormula:withDoc:at:on:">doUpdateForComplexFormula: aPvFormula withDoc: formulaSet at: timestamp on: timeSeries
	"Simulate a resync which includes a change of the PvFormula as refleced in the documentSource"

	| newFormula |
	self pvModel resyncAgent finalizeResync.
	newFormula := aPvFormula runCopy.
	newFormula
		resyncAction: #U;
		date: timestamp.

	"load the formula document with the same timestamp so that the lookup works properly"
	self pvModel formulaManager documentRegistry 
		registerFormulaSet: formulaSet 
		withTimestamp: timestamp.

	self pvModel formulaManager resyncFrom: newFormula.
	self pvModel resyncAgent commitResync.
	newFormula := self pvModel formulaManager itemWithId: newFormula id.
	queryInterface resyncFormulaArgumentsFor: newFormula.
	self pvModel resyncAgent commitResync.

	"We need to remove the heartbeat handler dependent for the old formula, since adding the time series will
		add a new dependency with the new version of the formula..."
	tupleRepository heartbeatHandlers do: [:each |
		each dependents asArray do: [:eachBroker |
			((eachBroker isKindOf: CMETimeSeries) and: [eachBroker formula isKindOf: LocalizedSchedule]) 
				ifFalse: 
					[eachBroker owner == aPvFormula formulaPrototype
						ifTrue: [each removeDependent: eachBroker]]]].

	" If the current formula changes, we need to re-shape it. We have to do this before we rebuild the rest because 
	code assumes time series inputs are consistent with formula arguments " 
	currentTimeSeries metricId = aPvFormula id
		ifTrue: 
			[self tupleRepository 
				atMsid: timeSeries msid 
				ifPresent: [:ts |
					self recompile: ts operator withFormula:  aPvFormula formulaPrototype]
				ifAbsent: []].
	self buildTimeSeriesFromCommand: timeSeries.</body>

<body package="PV-DC-Recomputations" selector="updateFormula:from:for:">updateFormula: originalFormula from: updatedFormula for: aTimeSeries
	"Update the given original formula with the information from the updated one. Note that for formulas, we have to handle both the formula iteself as well as the formula prototype."

	| documentSource formulaSet latestFormula |

	[self pvModel formulaManager resyncFrom: updatedFormula] 
		on: Error 
		do: [:ex | 
			" The current scheme is crazy and no matter which order you try to do it, it produces an error
			The formula a) needs to be in formula manager
				b) has groups loaded and
				c) has to have the doc loaded
			This would at least accomplish adding it to formula manager, then grouping can succeed
			"
		nil].

	queryInterface resyncFormulaRelationshipsFor: updatedFormula id.
	self pvModel formulaManager buildNameIndex.
	documentSource := updatedFormula sourceText.

	formulaSet := self pvModel formulaManager documentRegistry
		loadDocumentNamed: documentSource
		timestamp: updatedFormula resyncHistoricalDate
		ifMissing: [^self].
	
	self pvModel formulaManager resyncFrom: updatedFormula.
	self pvModel resyncAgent commitResync.

	latestFormula := self pvModel formulaManager itemWithId: updatedFormula id.
	self
		doUpdateForComplexFormula: latestFormula
		withDoc: formulaSet
		at: latestFormula timestamp
		on: aTimeSeries.</body>
</methods>

<methods>
<class-id>PROVISO.RecomputationEngine</class-id> <category>recomputation</category>

<body package="PV-DC-Recomputations" selector="computeInputsFor:">computeInputsFor: aTimeSeries
	"Determine what the inputs were to the formula the given time series represents."

	| wrapper toDelete |
	originalInputs := queryInterface 
		getLastInputsForTimeSeriesWithId: aTimeSeries msid 
		atOrBefore: scopeStartTimestamp.

	changedInputs := queryInterface
		getAllInputsForTimeSeriesWithId: aTimeSeries msid 
		between: scopeStartTimestamp
		and: scopeEndTimestamp.

	mergedInputs := SortedCollection sortBlock: [:a :b | 
		(a record resyncHistoricalDate &lt; b record resyncHistoricalDate) or: 
			[(a record resyncHistoricalDate = b record resyncHistoricalDate and: [a record inputTimeSeriesMsid &lt; b record inputTimeSeriesMsid])
				or: [a record resyncHistoricalDate = b record resyncHistoricalDate and: [a record inputTimeSeriesMsid = b record inputTimeSeriesMsid and: [(a record versionDate isNil or: [b record versionDate isNil]) or: [a record versionDate &lt; b record versionDate]]]]]].

	originalInputs do: [:each |
		wrapper := RecomputationAddInputWrapper
				on: each
				from: scopeStartTimestamp
				to: scopeEndTimestamp.
		mergedInputs add: wrapper].

	changedInputs do: [:each |
		wrapper := RecomputationAddInputWrapper
			on: each
			from: each timestamp
			to: scopeEndTimestamp.
		" Weed out plain dups here -- the queries above overlap by one second "
		(mergedInputs anySatisfy: [:eachInputWrapper | eachInputWrapper isSameAs: wrapper])
			ifFalse: [mergedInputs add: wrapper]].

	" Deal with input moves, etc "
	toDelete := OrderedCollection new.
	mergedInputs fold: [:prev :next |
		" Multiple simultaneous changes? "
		(prev record operatorTimeSeriesMsid = next record operatorTimeSeriesMsid
			and: [prev record inputTimeSeriesMsid = next record inputTimeSeriesMsid and: 
				[prev record resyncHistoricalDate = next record resyncHistoricalDate]])
			ifTrue: 
				[prev record versionDate = next record versionDate
					ifTrue: 
						[" Simultaneous change in the plan -- same version, must be time series move "
						prev record isActionDelete 
							ifTrue: [toDelete add: prev]
							ifFalse: 
								[next record isActionDelete 
									ifTrue: [toDelete add: next]]]
					ifFalse: [
						" next action is older version, leave just the last one "
						toDelete add: prev]
			].
		next
	].
	mergedInputs removeAll: toDelete.</body>

<body package="PV-DC-Recomputations" selector="computeMetadataChangesFor:">computeMetadataChangesFor: aTimeSeries
	"Determine the initial metadata setup for the given time series. That includes inputs, and the time series for those inputs. Also keep track of the time series separately, we use those to gather metric records."

	metadataChanges := SortedCollection
		with: (RecomputationAddTimeSeriesWrapper on: aTimeSeries from: scopeStartTimestamp to: scopeEndTimestamp).
	inputsByTime := OrderedCollection new.

	mergedInputs do: [:eachInputWrapper |
		| timeSeriesWrapper |
		timeSeriesWrapper := RecomputationAddTimeSeriesWrapper
			on: (queryInterface
				getLastTimeSeriesWithId: eachInputWrapper record inputTimeSeriesMsid
				atOrBefore: eachInputWrapper record timestamp)
			from: scopeStartTimestamp
			to: scopeEndTimestamp.
		inputsByTime add: timeSeriesWrapper.
		metadataChanges
			add: timeSeriesWrapper;
			addAll: 
				(self 
					computeMetadataChangeForMetricId: timeSeriesWrapper metricId 
					resourceId: timeSeriesWrapper resourceId 
					between: scopeStartTimestamp 
					and: timeSeriesWrapper timestamp);
			add: eachInputWrapper].</body>
</methods>


<methods>
<class-id>PROVISO.RecomputationMetadataChangeWrapper</class-id> <category>testing</category>

<body package="PV-DC-Recomputations" selector="isSameAs:">isSameAs: anotherWrapper

	^self record isSameAs: anotherWrapper record</body>
</methods>


<methods>
<class-id>PROVISO.RecomputationAddInputWrapper</class-id> <category>replay</category>

<body package="PV-DC-Recomputations" selector="replayFor:on:">replayFor: aRecomputationEngine on: aTupleRepository
	"Apply the receiver's record to the tuple repository during a recompute..."

	aRecomputationEngine buildInputFromCommand: self record.

	"We need to clean up spurious heartbeat brokers, because we are only interested
	in getting heartbeats to the current time series."
	aTupleRepository heartbeatHandlers do: [:each |
		each dependents asArray do: [:eachBroker |
			((eachBroker isKindOf: CMETimeSeries) and: [eachBroker formula isKindOf: LocalizedSchedule]) 
				ifFalse: 
					[eachBroker metricId = aRecomputationEngine currentTimeSeries metricId
						ifFalse: [each removeDependent: eachBroker]]]].</body>
</methods>


<methods>
<class-id>PROVISO.RecomputationQueryInterface</class-id> <category>queries</category>

<body package="PV-DC-Recomputations" selector="getAllInputsForTimeSeriesWithId:between:and:">getAllInputsForTimeSeriesWithId: anMsid between: scopeStartTimestamp and: scopeEndTimestamp
	" Answer all latest inputs for the given time series between the two timestamps.
	We scan the whole history looking for all inputs for this time series , then group by each input's msid."
	
	| inputs |
	inputs := OrderedCollection new.
	self dbSession
		readObjectsOfClass: PvInputBuildCommand
		with: (PvInputBuildCommand selectStatementForAllForMsid: anMsid between: scopeStartTimestamp and: scopeEndTimestamp)
		into: [:each | inputs add: each].

	^inputs</body>
</methods>


<methods>
<class-id>PROVISO.PvAbstractPlanBuildCommand</class-id> <category>testing</category>

<body package="PV-DM-Support" selector="isSameAs:">isSameAs: anotherCommand 
	
	^self resyncHistoricalDate = anotherCommand resyncHistoricalDate and: 
		[self resyncAction = anotherCommand resyncAction and: 
			[self versionDate = anotherCommand versionDate]]</body>
</methods>


<methods>
<class-id>PROVISO.PvTimeSeriesBuildCommand class</class-id> <category>instance creation</category>

<body package="PV-DM-Support" selector="fromTimeSeries:">fromTimeSeries: aTimeSeries

	| inst |
	inst := self new
		msid: aTimeSeries msid;
		metricId: aTimeSeries metricId;
		resourceId: aTimeSeries resourceId; 
		resyncHistoricalDate: ([aTimeSeries timestamp] on: Error do: [:ex | nil]);
		resyncAction: #I;
		streamLabelId: aTimeSeries streamLabel;
		isExplicitAsNumber: aTimeSeries isExplicit;
		requestId: aTimeSeries requestId.
	inst versionDate: inst timestamp.
	^inst</body>
</methods>


<methods>
<class-id>PROVISO.PvTimeSeriesBuildCommand</class-id> <category>printing</category>

<body package="PV-DM-Support" selector="printOn:">printOn: aStream

	aStream
		nextPutAll: self class name;
		nextPutAll: ' (';
		nextPutAll: self resyncAction asString;
		nextPutAll: ': ';
		nextPutAll: 'metricId: ';
		print: self metricId;
		nextPutAll: ' resourceId: ';
		print: self resourceId;
		nextPutAll: ' msid: ';
		print: self msid;
		nextPutAll: ' streamLabelId: ';
		print: self streamLabelId;
		nextPutAll: ' ts: ';
		nextPutAll: self timestamp pvOpportunisticPrintString;
		nextPutAll: ' version: ';
		nextPutAll: self versionDate pvOpportunisticPrintString;
		nextPut: $)</body>
</methods>


<methods>
<class-id>PROVISO.RecomputationAddTimeSeriesWrapper</class-id> <category>replay</category>

<body package="PV-DC-Recomputations" selector="replayFor:on:">replayFor: aRecomputationEngine on: aTupleRepository
	"Apply the receiver's record to the tuple repository during a recompute..."

	aRecomputationEngine buildTimeSeriesFromCommand: self record.

	"We need to clean up spurious heartbeat brokers, because we are only interested
	in getting heartbeats to the current time series."
	aTupleRepository heartbeatHandlers do: [:each |
		each dependents asArray do: [:eachBroker |
			((eachBroker isKindOf: CMETimeSeries) and: [eachBroker formula isKindOf: LocalizedSchedule]) 
				ifFalse: [	eachBroker metricId = aRecomputationEngine currentTimeSeries metricId
					ifFalse: [each removeDependent: eachBroker]]]].</body>
</methods>


<methods>
<class-id>PROVISO.PvInputBuildCommand class</class-id> <category>database support</category>

<body package="PV-DM-Support" selector="selectStatementForAllForMsid:between:and:">selectStatementForAllForMsid: operator between: startSeconds and: endSeconds 
	"self selectAllForMsid: 210098240 between: 1211218765 and: 1214870399"
	"This is the most generic select statement builder in our system.  In order to consolidate the entry point, we will encode special values for building different sql statements.
		If parametersDictionary doesn't include collectorList then we are selecting for all collectors
		if parametersDictionary doesn't include startSeconds, then we are doing a before date select"
	
	| realStartSeconds namedReplacements   |
	realStartSeconds := startSeconds.
	namedReplacements := Dictionary new.
	namedReplacements at: 'TABLE' put: self historyTableName asString.
	namedReplacements at: 'COLUMNLIST' put: (self allColumnsForSelectWithPrefix: 't1.').
	operator notNil
		ifTrue: [
			namedReplacements 
				at: 'OPERATORCLAUSE'
				put: 't1.idx_ts_operator = ', operator asString].

	endSeconds == Infinity positive
		ifTrue: [namedReplacements at: 'PRIMARYDATECONDITION' put: ' &gt;= ' , startSeconds asString , ' ']
		ifFalse: [
			startSeconds &gt; 0 
				ifTrue: [
					namedReplacements at: 'PRIMARYDATECONDITION' put: 
						' BETWEEN ' , realStartSeconds asString , ' AND ' , endSeconds asString , ' ']
				ifFalse: [namedReplacements at: 'PRIMARYDATECONDITION' put: ' &lt;= ' , endSeconds asString , ' ']].

	^'SELECT 
		%COLUMNLIST 
	FROM  %TABLE t1
	WHERE %OPERATORCLAUSE
		AND t1.dte_date %PRIMARYDATECONDITION ' namedBindWithArguments: namedReplacements</body>
</methods>


<methods>
<class-id>PROVISO.PvInputBuildCommand</class-id> <category>testing</category>

<body package="PV-DM-Support" selector="isSameAs:">isSameAs: anotherCommand 
	
	^(super isSameAs: anotherCommand) and: 
		[self operatorTimeSeriesMsid = anotherCommand operatorTimeSeriesMsid and: 
			[self inputTimeSeriesMsid = anotherCommand inputTimeSeriesMsid and: 
				[self operatorTimeSeriesLabel = anotherCommand operatorTimeSeriesLabel and: 
					[self sequenceNumber = anotherCommand sequenceNumber and: 
						[self currentResourceId = anotherCommand currentResourceId]]]]]</body>
</methods>


<methods>
<class-id>PROVISO.PvInputBuildCommand</class-id> <category>printing</category>

<body package="PV-DM-Support" selector="printOn:">printOn: aStream

	aStream
		nextPutAll: self class name;
		nextPutAll: ' (';
		nextPutAll: self resyncAction asString;
		nextPutAll: ': ';
		nextPutAll: 'operatorTimeSeriesMsid: ';
		print: self operatorTimeSeriesMsid;
		nextPutAll: ' operatorTimeSeriesLabel: ';
		print: self operatorTimeSeriesLabel;
		nextPutAll: ' inputTimeSeriesMsid: ';
		print: self inputTimeSeriesMsid;
		nextPutAll: ' currentResourceId: ';
		print: self currentResourceId;
		nextPutAll: ' sequenceNumber: ';
		print: self sequenceNumber;
		nextPutAll: ' ts: ';
		nextPutAll: self timestamp pvOpportunisticPrintString;
		nextPutAll: ' version: ';
		nextPutAll: self versionDate pvOpportunisticPrintString</body>
</methods>



</st-source>
