<?xml version="1.0"?>

<st-source>
<time-stamp>From VisualWorks®, 7.4.1 of May 30, 2006 on June 16, 2008 at 2:34:02 pm</time-stamp>


<methods>
<class-id>PROVISO.TupleStorage</class-id> <category>tuple queries - unioned</category>

<body package="PV-Tuple Storage Engine" selector="getTuplesForMid:rid:from:to:adjTime:count:dbSession:">getTuplesForMid: mid rid: rid from: startTime to: stopTime adjTime: adjTime count: count dbSession: dbSession
	"&lt;ReadStream&gt; Get all the tuples from tuples storage and the database for the given mid/rid.  Order by tuple timestamp (startTime may be greater than stopTime to indicate reverse order).  The startTime and endTime can be infinite to formulate a query that gets all tuples. The count is used to optimize the database queries, but does not restrict the number of tuples that can be in the resulting read stream (do not read more than count tuples from the resulting read stream, the results may not be accurate)."

	| agents transitionTime tupleStreams start stop tuplesExistInTupleStorage adjustmentStart adjustmentStop dbStart dbStop mergeSort direction needTuplesFromDB needTuplesFromTS |
	start := startTime ifNil: [Infinity negative].
	stop := stopTime ifNil: [Infinity positive].
	direction := self directionFrom: startTime to: stopTime ifIndeterminate: [^#() readStream].

	"Determine the earliest time for which TSE has tuples for the given mid/rid"
	agents := (self planAccessor labelsForMetricId: mid subelementId: rid between: startTime and: stopTime) 
				collect: [ :label | nil "self lookupNode: label ifAbsent: [nil]"].
	transitionTime := agents inject: Infinity negative into: [ :ts :ea | (ea notNil and: [ea earliestTimestamp &gt; ts]) ifTrue: [ea earliestTimestamp] ifFalse: [ts]].
	tuplesExistInTupleStorage := transitionTime ~~ Infinity negative.
	needTuplesFromDB := 
		tuplesExistInTupleStorage not or:
			[direction &gt;= 0
				ifTrue: [transitionTime &gt; start]
				ifFalse: [transitionTime &gt; stop]].
	needTuplesFromTS := 
		tuplesExistInTupleStorage and: 
			[direction &gt;= 0
				ifTrue: [transitionTime &lt;= stop]
				ifFalse: [transitionTime &lt;= start]].
	tupleStreams := OrderedCollection new.

	"We're going to shut off the TS queries for now..."
	needTuplesFromTS := false.
	needTuplesFromDB := true.

	"Now figure out the time ranges for the DB ajustment and tuple queries"
	needTuplesFromTS
		ifTrue: 
			[start &lt;= stop
				ifTrue:
					[adjustmentStart := transitionTime max: start.
					adjustmentStop := stop.
					dbStart := start.
					dbStop := transitionTime - 1]
				ifFalse: 
					[adjustmentStart := start.
					adjustmentStop := transitionTime max: stop.
					dbStart := transitionTime - 1.
					dbStop := stop]]
		ifFalse: 
			[dbStart := start.
			dbStop := stop].
	
	"1) First, lets go after the adjusted tuples from the database...do this first because we want the adjustments to shadow other results (if there are no tuples in tuple storage, no need to do this)."
	needTuplesFromTS ifTrue:
		[tupleStreams add:
			(self dbGetAdjustedSparseTuplesForMid: mid rid: rid from: adjustmentStart to: adjustmentStop adjTime: adjTime count: count dbSession: dbSession).

		"For now, we don't need the following because we don't support adjustments to periodic metrics"
		"tupleStreams add:
			(self dbGetAdjustedTuplesForMid: mid rid: rid from: adjustmentStart to: adjustmentStop adjTime: adjTime count: count dbSession: dbSession."].

	"2) Next go after tuples outside the Tuple Storage window in the DB"
	needTuplesFromDB ifTrue:
		[tupleStreams add:
			(self dbGetSparseTuplesForMid: mid rid: rid from: dbStart to: dbStop adjTime: adjTime count: count dbSession: dbSession).
		tupleStreams add:
			(self dbGetTuplesForMid: mid rid: rid from: dbStart to: dbStop adjTime: adjTime count: count dbSession: dbSession)].

	"3) Finally, get the tuples from tuple storage (if there aren't any tuples in tuple storage, skip this)"
	needTuplesFromTS 
		ifTrue:
			[tupleStreams add: (self tsGetTuplesForMid: mid rid: rid from: start to: stop)].

	direction &gt;= 0
		ifTrue: [mergeSort := MergeSortStream forTuplesOnAll: tupleStreams]
		ifFalse: [mergeSort := MergeSortStream forReverseTuplesOnAll: tupleStreams].
	^DeduplicatingStream 
		on: mergeSort
		comparisonBlock: [ :a :b | a timestamp = b timestamp]</body>
</methods>

</st-source>
